#!/bin/sh

# Fail script on error.
set -e

host_os=`uname -s`
echo $system

base_dir=/apps/rappture
build_dir=$base_dir/`date +%Y%m%d`
mkdir -p $build_dir

stage1_flags=""
stage2_flags=""
stage3_flags="--with-vtk"
rappture_flags=""

MAKE=make
case $host_os in 
   *Darwin* )
      DYLD_LIBRARY_PATH=$build_dir/lib
      export DYLD_LIBRARY_PATH
      ;;
   *Linux*  )
      LD_LIBRARY_PATH=$build_dir/lib
      export LD_LIBRARY_PATH
      ;;
   *FreeBSD* )
      MAKE=gmake
      LD_LIBRARY_PATH=$build_dir/lib
      export LD_LIBRARY_PATH
      ;;
esac

PATH=/usr/local/bin:/usr/bin:/bin:/usr/X11R6/bin:/usr/X11/bin:/usr/openwin/bin:$PATH:$build_dir/bin

export MAKE PATH

stage1() {
    pwd=`pwd`
    if test -d "stage1" ; then 
      cd stage1
    else 
      mkdir -p stage1
      cd stage1
      ../runtime/configure --prefix=$build_dir --exec_prefix=$build_dir \
       $stage1_flags
    fi
    ${MAKE} all
    ${MAKE} install
    cd $pwd
}

stage2() {
    pwd=`pwd`
    if test -d "stage2" ; then 
      cd stage2
    else 
      mkdir -p stage2
      cd stage2
      ../runtime/configure --prefix=$build_dir --exec_prefix=$build_dir \
       $stage2_flags
    fi
    ${MAKE} all
    ${MAKE} install
    cd $pwd
}


stage3() {
    pwd=`pwd`
    if test -d "stage3" ; then 
      cd stage3
    else 
      mkdir -p stage3
      cd stage3
      ../runtime/configure --prefix=$build_dir --exec_prefix=$build_dir \
       $stage3_flags
    fi
    ${MAKE} all
    ${MAKE} install
    cd $pwd
}

rappture() {
    pwd=`pwd`
    if test -d "stage.rappture" ; then 
      cd stage.rappture
    else 
      mkdir -p stage.rappture
      cd stage.rappture
      ../rappture/configure --prefix=$build_dir --exec_prefix=$build_dir --with-octave=/apps/octave/3.2.4/bin \
       $rappture_flags
    fi
    ${MAKE} all
    ${MAKE} install
    cd $pwd
}

stage1
stage2
stage3
rappture

exit 0
